'// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
'// ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO
'// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
'// PARTICULAR PURPOSE.
'//
'// Copyright (c) Microsoft Corporation. All rights reserved
'//
'// Translated to PowerBASIC
'// by Patrice Terrier <www.zapsolution.com>
'// on 07-26-2010.

#COMPILE EXE "TextAnimation.exe"
#DIM ALL
#INCLUDE ONCE "d2d1.inc"
#INCLUDE ONCE "d2d1Helper.inc"
#INCLUDE ONCE "Math.inc"
'#INCLUDE ONCE "wincodec.inc"
'#INCLUDE ONCE "dwrite.inc"
'#INCLUDE ONCE "d2dUtils.inc"

DECLARE FUNCTION zTrace LIB "zTrace.DLL" ALIAS "zTrace" (zMessage AS ASCIIZ) AS LONG

%AnimationStyle_None             = 0
%AnimationStyle_Translation      = 1
%AnimationStyle_Rotation         = 2
%AnimationStyle_Scaling          = 4

%TextRenderingMethod_Default     = 0
%TextRenderingMethod_Outline     = 1
%TextRenderingMethod_UseA8Target = 2
%TextRenderingMethod_NumValues   = 3

%maxElements                     = 10

GLOBAL g_pID2D1Helper AS ID2D1Helper
GLOBAL g_pD2DFactory AS ID2D1Factory
GLOBAL g_pTextFormat AS IDWriteTextFormat
GLOBAL g_pDWriteFactory AS IDWriteFactory
GLOBAL g_pRenderTarget AS ID2D1HwndRenderTarget
GLOBAL g_pBlackBrush AS ID2D1SolidColorBrush
GLOBAL g_pOpacityRT AS ID2D1BitmapRenderTarget
GLOBAL g_pTextLayout AS IDWriteTextLayout

GLOBAL g_overhangOffset AS D2D1_POINT_2F
GLOBAL g_startTime, g_animationStyle, g_renderingMethod  AS DWORD

FUNCTION WinMain (BYVAL hInstance AS DWORD, BYVAL hPrevInstance AS DWORD, BYVAL lpszCmdLine AS ASCIIZ PTR, BYVAL nCmdShow AS LONG) AS LONG

    LOCAL hWndMain    AS DWORD
    LOCAL Msg         AS tagMsg
    LOCAL wc          AS WNDCLASSEX
    LOCAL szClassName AS ASCIIZ * 80
    LOCAL rc          AS RECT
    LOCAL szCaption   AS ASCIIZ * 255
    LOCAL hr, IsInitialized AS LONG
    LOCAL x, y, w, h  AS LONG

    ' // Create an instance of the CD2D1Helper class
    g_pID2D1Helper = CLASS "CD2D1Helper"
    IF ISNOTHING(g_pID2D1Helper) THEN
       MsgBox "Unable to create the D2D1Helper class"
       EXIT FUNCTION
    END IF

    hr = CreateDeviceIndependentResources()
    IF hr = 0 THEN

       szClassName = "ZD2D1TEXT"

       IsInitialized = GetClassInfoEx(hInstance, szClassName, wc)
       IF IsInitialized   = 0 THEN
          wc.cbSize        = SIZEOF(wc)
          wc.style         = %CS_HREDRAW OR %CS_VREDRAW
          wc.lpfnWndProc   = CODEPTR(WndProc)
          wc.cbClsExtra    = 0
          wc.cbWndExtra    = 0
          wc.hInstance     = hInstance
          wc.hIcon         = LoadIcon (%NULL, BYVAL %IDI_APPLICATION)
          wc.hCursor       = LoadCursor(%NULL, BYVAL %IDC_ARROW)
          wc.hbrBackground = %NULL ' GetStockObject(%BLACK_BRUSH)
          wc.lpszMenuName  = %NULL
          wc.lpszClassName = VARPTR(szClassName)
          wc.hIconSm       = wc.hIcon
          IF RegisterClassEx(wc) THEN IsInitialized = %TRUE
       END IF

       IF IsInitialized THEN
        ' // Create the application window.
        ' //
        ' // Because the CreateWindow function takes its size in pixels, we
        ' // obtain the system DPI and use it to scale the window size.
          LOCAL dpiX, dpiY AS SINGLE
          g_pD2DFactory.GetDesktopDpi(dpiX, dpiY)

          ' Window caption
          szCaption = "Direct2D Demo App"

          w = CEIL(640 * dpiX / 96)
          h = CEIL(480 * dpiY / 96)
          x = MAX&((GetSystemMetrics(%SM_CXSCREEN) - w) \ 2, 0)
          y = MAX&((GetSystemMetrics(%SM_CYSCREEN) - h) \ 2, 0)

          ' Create a window using the registered class
          hWndMain = CreateWindowEx(%WS_EX_CONTROLPARENT, _    ' extended style
                             szClassName, _                    ' window class name
                             szCaption, _                      ' window caption
                             %WS_OVERLAPPEDWINDOW, _           ' window styles
                             x, _                              ' initial x position
                             y, _                              ' initial y position
                             w, _                              ' initial width
                             h, _                              ' initial height
                             %NULL, _                          ' parent window handle
                             0, _                              ' window menu handle
                             hInstance, _                      ' program instance handle
                             BYVAL %NULL)                      ' creation parameters

          IF hWndMain THEN

             '// Initialize global data
             g_startTime       = 0
             g_animationStyle  = %AnimationStyle_Translation OR %AnimationStyle_Rotation OR %AnimationStyle_Scaling
             g_renderingMethod = %TextRenderingMethod_Default

           ' Show the main window
             CALL ShowWindow(hWndMain, nCmdShow)
             'CALL SetForegroundWindow(hWndMain)
             CALL Updatewindow(hWndMain)

             WHILE GetMessage(Msg, %NULL, 0, 0)
                 IF IsDialogMessage(hWndMain, Msg) = %FALSE THEN
                    CALL TranslateMessage(msg)                ' Translate The Message
                    CALL DispatchMessage(msg)                 ' Dispatch The Message
                 END IF
             WEND

             FUNCTION = msg.wParam

          END IF
       END IF
    END IF

'   // Cleanup
    g_pID2D1Helper   = NOTHING
    g_pD2DFactory    = NOTHING
    g_pTextFormat    = NOTHING
    g_pDWriteFactory = NOTHING
    g_pBlackBrush    = NOTHING
    g_pOpacityRT     = NOTHING
    g_pRenderTarget  = NOTHING

END FUNCTION

'//  This function is used to create resources which are not bound
'//  to any device. Their lifetime effectively extends for the
'//  duration of the app. These resources include the D2D,
'//  DWrite factories; and a DWrite Text Format object
'//  (used for identifying particular font characteristics) and
'//  a D2D geometry.
FUNCTION CreateDeviceIndependentResources() AS LONG

    LOCAL hr AS LONG, stringLength AS DWORD
    LOCAL msc_fontName, msc_localName, sc_helloWorld AS WSTRINGZ * 64
    LOCAL msc_fontSize AS SINGLE

    msc_fontName = "Gabriola"
    msc_localName = ""
    sc_helloWorld = "The quick brown fox jumped over the lazy dog!"
    msc_fontSize = 50
    stringLength = LEN(sc_helloWorld)

    '//create D2D factory
    hr = D2D1CreateFactory(%D2D1_FACTORY_TYPE_SINGLE_THREADED, $IID_ID2D1Factory, BYVAL %NULL, g_pD2DFactory)
    IF SUCCEEDED(hr) THEN
        '//create DWrite factory
        hr = DWriteCreateFactory( _
            %DWRITE_FACTORY_TYPE_ISOLATED, _ '//DWRITE_FACTORY_TYPE_SHARED
            $IID_IDWriteFactory, _           '//__uuidof(g_pDWriteFactory), _
            g_pDWriteFactory _
            )
    END IF

    IF SUCCEEDED(hr) THEN
        '//create DWrite text format object
        hr = g_pDWriteFactory.CreateTextFormat( _
            msc_fontName, _
            NOTHING, _
            %DWRITE_FONT_WEIGHT_NORMAL, _
            %DWRITE_FONT_STYLE_NORMAL, _
            %DWRITE_FONT_STRETCH_NORMAL, _
            msc_fontSize, _
            msc_localName, _ // L"", _ '//locale
            g_pTextFormat _
            )
    END IF

    IF SUCCEEDED(hr) THEN
        '//center the text horizontally
        g_pTextFormat.SetTextAlignment(%DWRITE_TEXT_ALIGNMENT_CENTER)

        hr = g_pDWriteFactory.CreateTextLayout( _
            sc_helloWorld, _
            stringLength, _
            g_pTextFormat, _
            300, _  '// maxWidth
            1000, _ '// maxHeight
            g_pTextLayout _
            )
    END IF

    IF SUCCEEDED(hr) THEN
        '//
        '// We use typographic features here to show how to account for the
        '// overhangs that these features will produce. See the code in
        '// ResetAnimation that calls GetOverhangMetrics(). Note that there are
        '// fonts that can produce overhangs even without the use of typographic
        '// features- this is just one example.
        '//
        LOCAL pTypography AS IDWriteTypography
        hr = g_pDWriteFactory.CreateTypography(pTypography)
        IF SUCCEEDED(hr) THEN
            LOCAL fontFeature AS DWRITE_FONT_FEATURE
            fontFeature.nameTag = %DWRITE_FONT_FEATURE_TAG_STYLISTIC_SET_7
            fontFeature.parameter = 1
            hr = pTypography.AddFontFeature(fontFeature)
            IF SUCCEEDED(hr) THEN
               LOCAL textRange AS DWRITE_TEXT_RANGE
               textRange.startPosition = 0
               textRange.length = stringLength
               hr = g_pTextLayout.SetTypography(pTypography, textRange)
            END IF

            pTypography = NOTHING
        END IF
    END IF

    FUNCTION = hr

END FUNCTION

'// This function creates resources which are bound to a particular
'// D3D device. It's all centralized here, in case the resources
'// need to be recreated in case of D3D device loss (eg. display
'// change, remoting, removal of video card, etc).
FUNCTION CreateDeviceResources(BYVAL hWnd AS LONG) AS LONG

    LOCAL hr AS LONG

    IF ISNOTHING(g_pRenderTarget) THEN

        LOCAL rc AS RECT
        CALL GetClientRect(hWnd, rc)

        LOCAL uSize AS D2D1_SIZE_U
        uSize = g_pID2D1Helper.SizeU(rc.nRight - rc.nLeft, rc.nBottom - rc.nTop)

        '// Create a D2D render target
        '//
        '// Note: we only use D2D1_PRESENT_OPTIONS_IMMEDIATELY so that we can
        '// easily measure the framerate. Most apps should not use this
        '// flag.
        '//
        hr = g_pD2DFactory.CreateHwndRenderTarget( _
             g_pID2D1Helper.RenderTargetProperties(), _
             g_pID2D1Helper.HwndRenderTargetProperties(hWnd, uSize, %D2D1_PRESENT_OPTIONS_IMMEDIATELY), _
             g_pRenderTarget _
             )
        IF SUCCEEDED(hr) THEN
            '//
            '// Nothing in this sample requires antialiasing so we set the antialias
            '// mode to aliased up front.
            '//
            g_pRenderTarget.SetAntialiasMode(%D2D1_ANTIALIAS_MODE_ALIASED)

            '//create a black brush
            LOCAL brushProperties AS D2D1_BRUSH_PROPERTIES
            brushProperties.opacity = 1.0
            brushProperties.transform =  g_pID2D1Helper.IdentityMatrix()
            hr = g_pRenderTarget.CreateSolidColorBrush( _
                 g_pID2D1Helper.ColorF_2(%D2D1_Black, 1.0), _
                 brushProperties, _
                 g_pBlackBrush)
        END IF
        IF SUCCEEDED(hr) THEN
            hr = ResetAnimation(%TRUE)
        END IF
    END IF

    FUNCTION = hr

END FUNCTION


'//  Discard device-specific resources which need to be recreated
'//  when a D3D device is lost
SUB DiscardDeviceResources()
    g_pRenderTarget = NOTHING
    g_pBlackBrush   = NOTHING
    g_pOpacityRT    = NOTHING
END SUB

'// Responds to input from the user.
FUNCTION OnChar(BYVAL Charkey AS LONG) AS LONG

    LOCAL hr, bResetAnimation, bResetClock AS LONG

    bResetAnimation = -1
    bResetClock = -1

    SELECT CASE LONG Charkey

    CASE %VK_T
        IF (g_animationStyle AND %AnimationStyle_Translation) THEN
            g_animationStyle = g_animationStyle AND (NOT %AnimationStyle_Translation)
        ELSE
            g_animationStyle = g_animationStyle OR %AnimationStyle_Translation
        END IF

    CASE %VK_R
        IF (g_animationStyle AND %AnimationStyle_Rotation) THEN
            g_animationStyle = g_animationStyle AND (NOT %AnimationStyle_Rotation)
        ELSE
            g_animationStyle = g_animationStyle OR %AnimationStyle_Rotation
        END IF

    CASE %VK_S
        IF (g_animationStyle AND %AnimationStyle_Scaling) THEN
            g_animationStyle = g_animationStyle AND (NOT %AnimationStyle_Scaling)
        ELSE
            g_animationStyle = g_animationStyle OR %AnimationStyle_Scaling
        END IF

    CASE %VK_1, %VK_2, %VK_3
        g_renderingMethod = CharKey - 49 '// static_cast<%TextRenderingMethod_Enum>(key - '1')
        bResetClock = 0

    CASE ELSE
        bResetAnimation = 0
        bResetClock = 0
    END SELECT

    IF bResetAnimation THEN
        hr = ResetAnimation(bResetClock)
    END IF

    FUNCTION = hr

END FUNCTION

'// This procedure updates the window title bar with info about the
'// current animation style and rendering method. It also outputs
'// the framerate.
SUB UpdateWindowText(BYVAL hWnd AS LONG)

    LOCAL sTitle, sT, sR, sS, sMethod, sFps AS STRING

    STATIC sWasTitle AS STRING
    STATIC dwFrames, dwCurrentTime, dwLastUpdateTime, dwElapsedTime AS DWORD

    dwFrames += 1
    dwCurrentTime = timeGetTime()
    dwElapsedTime = dwCurrentTime - dwLastUpdateTime
    IF dwElapsedTime > 999 THEN

        IF dwLastUpdateTime THEN sFps = "," + STR$( dwFrames * 1000 \ dwElapsedTime) + " fps"
        dwFrames = 0: dwLastUpdateTime = dwCurrentTime  

'    STATIC sc_lastTimeStatusShown AS QUAD
'
'    '//
'    '// Update the window status no more than 10 times a second. Without this
'    '// check, the performance bottleneck could potentially be the time it takes
'    '// for Windows to update the title.
'    '//
'    IF  g_times.GetCount() > 0 AND g_times.GetLast() > sc_lastTimeStatusShown + 1000000 THEN
'        '// Determine the frame rate by computing the difference in clock time
'        '// between this frame and the frame we rendered 10 frames ago.
'        sc_lastTimeStatusShown = g_times.GetLast()
'
'        LOCAL qfrequency AS QUAD
'        QueryPerformanceFrequency(qfrequency)
'
'        IF (g_times.GetCount() > 0) THEN
'            sFps = "," + STR$((g_times.GetCount()-1) * frequency.QuadPart /
'                    static_cast<float>((g_times.GetLast() - g_times.GetFirst()))) + " fps"
'        END IF

        SELECT CASE LONG g_renderingMethod
        CASE %TextRenderingMethod_Default
            sMethod = "Default"
        CASE %TextRenderingMethod_Outline
            sMethod = "Outline"
        CASE %TextRenderingMethod_UseA8Target
            sMethod = "UseA8Target"
        END SELECT

        sTitle = "AnimationStyle: "
        IF (g_animationStyle AND %AnimationStyle_Translation) THEN sT = "+t" ELSE sT = "-t"
        IF (g_animationStyle AND %AnimationStyle_Rotation) THEN sR = "+r" ELSE sR = "-r"
        IF (g_animationStyle AND %AnimationStyle_Scaling) THEN sS = "+s" ELSE sS = "-s"
        sTitle = BUILD$(sTitle, sT, sR, sS, ", Method: ", sMethod, sFps, $NUL)

        IF sTitle <> sWasTitle THEN 
           SetWindowText(hWnd, BYVAL STRPTR(sTitle))
           sWasTitle = sTitle
        END IF
    END IF
END SUB

'// This function does the necessary work to change the current
'// animation style.
FUNCTION ResetAnimation(BYVAL resetClock AS LONG) AS LONG

    LOCAL hr AS LONG

    IF resetClock THEN g_startTime = GetTickCount()

    '// Release the opacity mask. We will regenerate it if the current animation
    '// style demands it.
    g_pOpacityRT = NOTHING

    IF g_renderingMethod = %TextRenderingMethod_Outline THEN

        '// Set the rendering mode to OUTLINE mode. To do this we first create
        '// a default params object and then make a copy with the given modification.

        LOCAL pDefaultParams AS IDWriteRenderingParams
        hr = g_pDWriteFactory.CreateRenderingParams(pDefaultParams)

        IF SUCCEEDED(hr) THEN
            LOCAL pRenderingParams AS IDWriteRenderingParams
            hr = g_pDWriteFactory.CreateCustomRenderingParams( _
                pDefaultParams.GetGamma(), _
                pDefaultParams.GetEnhancedContrast(), _
                pDefaultParams.GetClearTypeLevel(), _
                pDefaultParams.GetPixelGeometry(), _
                %DWRITE_RENDERING_MODE_OUTLINE, _
                pRenderingParams _
                )
            IF SUCCEEDED(hr) THEN
                g_pRenderTarget.SetTextRenderingParams(pRenderingParams)

                pRenderingParams = NOTHING
            END IF
            pDefaultParams = NOTHING
        END IF
    ELSE
        '// Reset the rendering mode to default.
        g_pRenderTarget.SetTextRenderingParams(NOTHING)
    END IF

    IF SUCCEEDED(hr) AND g_renderingMethod = %TextRenderingMethod_UseA8Target THEN

        '// Create a compatible A8 Target to store the text as an opacity mask.
        '//
        '// Note: To reduce sampling error in the scale animation, it might be
        '//       preferable to create multiple masks for the text at different
        '//       resolutions.

        LOCAL dpiX, dpiY AS SINGLE
        g_pRenderTarget.GetDpi(dpiX, dpiY)

        '// It is important to obtain the overhang metrics here in case the text
        '// extends beyond the layout max-width and max-height.
        LOCAL overhangMetrics AS DWRITE_OVERHANG_METRICS
        g_pTextLayout.GetOverhangMetrics(overhangMetrics)

        '// Because the overhang metrics can be off slightly given that these
        '// metrics do not account for antialiasing, we add an extra pixel for
        '// padding.
        LOCAL padding AS D2D1_SIZE_F
        padding = g_pID2D1Helper.SizeF(96.0 / dpiX, 96.0 / dpiY)
        g_overhangOffset = g_pID2D1Helper.Point2F(CEIL(overhangMetrics.left + padding.width), CEIL(overhangMetrics.top + padding.height))

        '// The true width of the text is the max width + the overhang
        '// metrics + padding in each direction.
        LOCAL maskSize AS D2D1_SIZE_F
        maskSize = g_pID2D1Helper.SizeF( _
            overhangMetrics.right + padding.width + g_overhangOffset.x + g_pTextLayout.GetMaxWidth(), _
            overhangMetrics.bottom + padding.height + g_overhangOffset.y + g_pTextLayout.GetMaxHeight() _
            )

        '// Round up to the nearest pixel
        LOCAL maskPixelSize AS D2D1_SIZE_U
        maskPixelSize = g_pID2D1Helper.SizeU( _
            CEIL(maskSize.width * dpiX / 96.0), _
            CEIL(maskSize.height * dpiY / 96.0) _
            )

        '// Create the compatible render target using desiredPixelSize to avoid
        '// blurriness issues caused by a fractional-pixel desiredSize.
        LOCAL alphaOnlyFormat AS D2D1_PIXEL_FORMAT
        alphaOnlyFormat = g_pID2D1Helper.PixelFormat(%DXGI_FORMAT_A8_UNORM, %D2D1_ALPHA_MODE_PREMULTIPLIED)
        hr = g_pRenderTarget.CreateCompatibleRenderTarget( _
                BYVAL %NULL, _
                maskPixelSize, _
                alphaOnlyFormat, _
                %D2D1_COMPATIBLE_RENDER_TARGET_OPTIONS_NONE, _
                g_pOpacityRT _
                )

        IF SUCCEEDED(hr) THEN
            '//
            '// Draw the text to the opacity mask. Note that we can use pixel
            '// snapping now given that subpixel translation can now happen during
            '// the FillOpacityMask method.
            '//
            g_pOpacityRT.BeginDraw()
            g_pOpacityRT.Clear(g_pID2D1Helper.ColorF_2(%D2D1_Black, 0.0))
            g_pOpacityRT.DrawTextLayout( _
                g_overhangOffset, _
                g_pTextLayout, _
                g_pBlackBrush, _
                %D2D1_DRAW_TEXT_OPTIONS_NO_SNAP _
                )
            hr = g_pOpacityRT.EndDraw()
        END IF
    END IF

    FUNCTION = hr

END FUNCTION

'//  Calculates the transform based on the current time
SUB CalculateTransform(pTransform AS D2D1_MATRIX_3X2_F)

    '// calculate a 't' value that will linearly interpolate from 0 to 1 and back every 20 seconds
    LOCAL currentTime AS DWORD
    currentTime = GetTickCount()
    IF g_startTime = 0 THEN
        g_startTime = currentTime
    END IF
    LOCAL t AS SINGLE
    t = 2 * (( currentTime - g_startTime) MOD 20000) / 20000.0
    IF t > 1.0 THEN
        t = 2 - t
    END IF

    '// calculate animation parameters
    LOCAL rotation, translationOffset, scaleMultiplier AS SINGLE
    rotation = 0
    translationOffset = 0
    scaleMultiplier = 1.0
    IF (g_animationStyle AND %AnimationStyle_Translation) THEN
        '// range from -100 to 100
        translationOffset = (t - 0.5) * 200
    END IF

    IF (g_animationStyle AND %AnimationStyle_Rotation) THEN
        '// range from 0 to 360
        rotation = t * 360.0
    END IF

    IF (g_animationStyle AND %AnimationStyle_Scaling) THEN
        '// range from 1/4 to 2x the normal size
        scaleMultiplier = t * 1.75 + 0.25
    END IF

    LOCAL renderTargetSize AS D2D1_SIZE_F
    renderTargetSize = g_pRenderTarget.GetSize()

    LOCAL center AS D2D1_POINT_2F

    '// *pTransform =
    '//      D2D1::Matrix3x2F::Rotation(rotation)
    '//    * D2D1::Matrix3x2F::Scale(scaleMultiplier, scaleMultiplier)
    '//    * D2D1::Matrix3x2F::Translation(translationOffset + size.width / 2.0f, translationOffset + size.height / 2.0f);   
    pTransform = g_pID2D1Helper.MultiplyMatrices(g_pID2D1Helper.MatrixRotation(rotation, center), _
                 g_pID2D1Helper.MatrixScale(scaleMultiplier, scaleMultiplier))
    pTransform = g_pID2D1Helper.MultiplyMatrices(pTransform, _
                 g_pID2D1Helper.MatrixTranslation(translationOffset + renderTargetSize.width / 2.0, translationOffset + renderTargetSize.height / 2.0))

END SUB

'//  Note that this function will not render anything if the window
'//  is occluded (e.g. when the screen is locked).
'//  Also, this function will automatically discard device-specific
'//  resources if the D3D device disappears during function
'//  invocation, and will recreate the resources the next time it's
'//  invoked.
FUNCTION OnRender(BYVAL hWnd AS LONG) AS LONG

    LOCAL hr AS LONG

    '//
    '// We use a ring buffer to store the clock time for the last 10 frames.
    '// This lets us eliminate a lot of noise when computing framerate.
    '//
'    LOCAL qTime AS QUAD
'    QueryPerformanceCounter(qTime)
'    g_times.Add(time.QuadPart)

    hr = CreateDeviceResources(hWnd)

    IF SUCCEEDED(hr) THEN
       IF (g_pRenderTarget.CheckWindowState() AND %D2D1_WINDOW_STATE_OCCLUDED) = 0 THEN

           LOCAL transform AS D2D1_MATRIX_3X2_F
           CalculateTransform(transform)

           g_pRenderTarget.BeginDraw()

           g_pRenderTarget.Clear(g_pID2D1Helper.ColorF_2(%D2D1_White, 1.0))

           g_pRenderTarget.SetTransform(transform)

           LOCAL textMetrics AS DWRITE_TEXT_METRICS
           g_pTextLayout.GetMetrics(textMetrics)

           IF (g_renderingMethod = %TextRenderingMethod_UseA8Target) THEN

               '// Offset the destination rect such that the text will be centered
               '// on the render target. Given that we have offset the text in the
               '// A8 target by the overhang offset, we must factor that into the
               '// destination rect now.
               LOCAL opacityRTSize AS D2D1_SIZE_F
               opacityRTSize = g_pOpacityRT.GetSize()
               LOCAL offset AS D2D1_POINT_2F
               offset = g_pID2D1Helper.Point2F( _
                       -textMetrics.width / 2.0 - g_overhangOffset.x, _
                       -textMetrics.height / 2.0 - g_overhangOffset.y _
                       )

               '// Round the offset to the nearest pixel. Note that the rounding
               '// done here is unecessary, but it causes the text to be less
               '// blurry.
               LOCAL dpiX, dpiY AS SINGLE
               g_pRenderTarget.GetDpi(dpiX, dpiY)
               LOCAL roundedOffset AS D2D1_POINT_2F
               roundedOffset = g_pID2D1Helper.Point2F( _
                   floor_(offset.x * dpiX / 96.0 + 0.5) * 96.0 / dpiX, _
                   floor_(offset.y * dpiY / 96.0 + 0.5) * 96.0 / dpiY _
                   )

               LOCAL destinationRect AS D2D1_RECT_F
               destinationRect = g_pID2D1Helper.RectF( _
                   roundedOffset.x, _
                   roundedOffset.y, _
                   roundedOffset.x + opacityRTSize.width, _
                   roundedOffset.y + opacityRTSize.height _
                   )

               LOCAL pBitmap AS ID2D1Bitmap
               g_pOpacityRT.GetBitmap(pBitmap)

               pBitmap.GetDpi(dpiX, dpiY)

               '// The antialias mode must be set to D2D1_ANTIALIAS_MODE_ALIASED
               '// for this method to succeed. We've set this mode already though
               '// so no need to do it again.
               g_pRenderTarget.FillOpacityMask( _
                   pBitmap, _
                   g_pBlackBrush, _
                   %D2D1_OPACITY_MASK_CONTENT_TEXT_NATURAL, _
                   destinationRect _
                   )

               pBitmap = NOTHING

           ELSE

               LOCAL origin AS D2D1_POINT_2F
               origin = g_pID2D1Helper.Point2F(-textMetrics.width / 2.0, -textMetrics.height / 2.0)
               '// Disable pixel snapping to get a smoother animation.
               g_pRenderTarget.DrawTextLayout( _
                   origin, _
                   g_pTextLayout, _
                   g_pBlackBrush, _
                   %D2D1_DRAW_TEXT_OPTIONS_NO_SNAP _
                   )

           END IF

           hr = g_pRenderTarget.EndDraw()

           IF hr = %D2DERR_RECREATE_TARGET THEN
               hr = 0
               DiscardDeviceResources()
           END IF
           '// To animate as quickly as possible, we request another WM_PAINT
           '// immediately.
           CALL InvalidateRect(hWnd, BYVAL %NULL, %FALSE)

        END IF

    END IF

    UpdateWindowText(hWnd)

    FUNCTION = hr
END FUNCTION

'//  If the application receives a WM_SIZE message, this method
'//  resize the render target appropriately.
SUB OnResize(BYVAL nWidth AS LONG, BYVAL nHeight AS LONG)
    IF NOT ISNOTHING(g_pRenderTarget) THEN
       LOCAL su AS D2D1_SIZE_U

       su.width = nWidth
       su.height = nHeight
       '// Note: This method can fail, but it's okay to ignore the
       '// error here -- it will be repeated on the next call to
       '// EndDraw.
       g_pRenderTarget.Resize(su)
    END IF
END SUB

'// The window message handler.
FUNCTION WndProc(BYVAL hWnd AS LONG, BYVAL Msg AS LONG, BYVAL wParam AS LONG, BYVAL lParam AS LONG) AS LONG

    LOCAL ps AS PAINTSTRUCT

    SELECT CASE LONG Msg

    CASE %WM_SIZE
         OnResize(LO(INTEGER, lParam), HI(INTEGER, lParam))

    CASE %WM_KEYDOWN
         CALL OnChar(wParam)
         FUNCTION = 0: EXIT FUNCTION

    CASE %WM_PAINT, %WM_DISPLAYCHANGE
         CALL BeginPaint(hWnd, ps)
         CALL OnRender(hWnd)
         CALL EndPaint(hWnd, ps)
         FUNCTION = 0: EXIT FUNCTION

    CASE %WM_DESTROY
         CALL PostQuitMessage(0)
         FUNCTION = 0: EXIT FUNCTION
    END SELECT

    FUNCTION = DefWindowProc(hWnd, Msg, wParam, lParam)

END FUNCTION

